from .qutie import QtWidgets

from .widget import BaseWidget

__all__ = ['Frame']

class Frame(BaseWidget):

    QtClass = QtWidgets.QFrame

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



from .qutie import QtWidgets
from .qutie import QutieStub

from .widget import BaseWidget

__all__ = ['ComboBox']

class ComboBox(BaseWidget):

    QtClass = QtWidgets.QComboBox

    def __init__(self, items=None, *, current=None, editable=False,
                 changed=None, **kwargs):
        super().__init__(**kwargs)
        if items is not None:
            self.items = items
        if current is not None:
            self.current = current
        self.editable = editable
        self.changed = changed
        # Connect signals
        self.qt.currentIndexChanged.connect(self.__handle_changed)

    @property
    def items(self):
        return list(self)

    @items.setter
    def items(self, items):
        self.clear()
        for item in items:
            self.append(item)

    def clear(self):
        self.qt.clear()

    def append(self, value):
        self.qt.addItem(format(value), value)

    def insert(self, index, value):
        if index < 0:
            index = max(0, len(self) + index)
        self.qt.insertItem(index, format(value), value)

    def remove(self, value):
        self.qt.removeItem(self.qt.findData(value))

    @property
    def current(self):
        return self.qt.itemData(self.qt.currentIndex())

    @current.setter
    def current(self, item):
        index = self.qt.findData(item)
        self.qt.setCurrentIndex(index)

    def index(self, item):
        index = self.qt.findData(item)
        if index < 0:
            raise ValueError("item not in list")
        return index

    @property
    def editable(self):
        return self.qt.isEditable()

    @editable.setter
    def editable(self, value):
        self.qt.setEditable(value)

    @property
    def changed(self):
        return self.__changed

    @changed.setter
    def changed(self, changed):
        self.__changed = changed

    def __handle_changed(self, index):
        if callable(self.changed):
            value = self.items[index]
            self.changed(value)

    def __getitem__(self, key):
        return self.qt.itemData(key)

    def __setitem__(self, key, value):
        self.qt.setItemText(key, format(value))
        self.qt.setItemData(key, value)

    def __delitem__(self, key):
        self.qt.removeItem(key)

    def __len__(self):
        return self.qt.count()

    def __iter__(self):
        return (self[row] for row in range(len(self)))



from .qutie import QtWidgets

from .widget import BaseWidget

__all__ = ['Column', 'Row', 'Spacer']

class BoxLayout(BaseWidget):

    QtLayoutClass = QtWidgets.QBoxLayout

    def __init__(self, *children, stretch=None, **kwargs):
        super().__init__(**kwargs)
        self.qt.setLayout(self.QtLayoutClass())
        self.qt.layout().setContentsMargins(0, 0, 0, 0)
        for child in children:
            self.append(child)
        if stretch is not None:
            self.stretch = stretch

    @property
    def children(self):
        return list(self)

    @property
    def stretch(self):
        value = []
        for index in range(len(self)):
            value.append(self.qt.layout().stretch(index))
        return value

    @stretch.setter
    def stretch(self, value):
        value = list(value) + ([0] * (len(self) - len(value)))
        for index in range(len(self)):
            self.qt.layout().setStretch(index, value[index])

    def append(self, child):
        self.qt.layout().addWidget(child.qt)

    def insert(self, index, child):
        self.qt.layout().insertWidget(index, child.qt)

    def remove(self, child):
        index = self.qt.layout().indexOf(child.qt)
        self.qt.layout().takeAt(index)

    def __getitem__(self, key):
        return self.qt.layout().itemAt(key).widget().reflection()

    def __len__(self):
        return self.qt.layout().count()

    def __iter__(self):
        return (self[index] for index in range(len(self)))

class Column(BoxLayout):

    QtLayoutClass = QtWidgets.QVBoxLayout

class Row(BoxLayout):

    QtLayoutClass = QtWidgets.QHBoxLayout

class Spacer(BaseWidget):

    QtSizePolicy = {
        'fixed': QtWidgets.QSizePolicy.Fixed,
        'minimum': QtWidgets.QSizePolicy.Minimum,
        'maximum': QtWidgets.QSizePolicy.Maximum,
        'preferred': QtWidgets.QSizePolicy.Preferred,
        'expanding': QtWidgets.QSizePolicy.Expanding,
        'minimum_expanding': QtWidgets.QSizePolicy.MinimumExpanding,
        'ignored': QtWidgets.QSizePolicy.Ignored
    }

    def __init__(self, horizontal=True, vertical=True, **kwargs):
        super().__init__(**kwargs)
        self.qt.setSizePolicy(
            self.QtSizePolicy.get('expanding' if horizontal else 'fixed'),
            self.QtSizePolicy.get('expanding' if vertical else 'fixed')
        )



import signal
import sys

from .qutie import QtCore
from .qutie import QtGui
from .qutie import QtWidgets

from .icon import Icon
from .object import Object

__all__ = ['CoreApplication', 'GuiApplication', 'Application']

class CoreApplication(Object):

    QtClass = QtCore.QCoreApplication

    def __init__(self, name=None, *, version=None, organization=None,
                 domain=None, name_changed=None, version_changed=None,
                 organization_changed=None, domain_changed=None, **kwargs):
        super().__init__(qt=(sys.argv,), **kwargs)
        if name is not None:
            self.name = name
        if version is not None:
            self.version = version
        if organization is not None:
            self.organization = organization
        if domain is not None:
            self.domain = domain
        self.name_changed = name_changed
        self.version_changed = version_changed
        self.organization_changed = organization_changed
        self.domain_changed = domain_changed
        # Connect signals
        self.qt.applicationNameChanged.connect(self.__handle_name_changed)
        self.qt.applicationVersionChanged.connect(self.__handle_version_changed)
        self.qt.organizationNameChanged.connect(self.__handle_organization_changed)
        self.qt.organizationDomainChanged.connect(self.__handle_domain_changed)

    @classmethod
    def instance(cls):
        if cls.QtClass.instance() is not None:
            return cls.QtClass.instance().reflection()
        return None

    @property
    def name(self):
        return self.qt.applicationName()

    @name.setter
    def name(self, value):
        self.qt.setApplicationName(value)

    @property
    def version(self):
        return self.qt.applicationVersion()

    @version.setter
    def version(self, value):
        self.qt.setApplicationVersion(value)

    @property
    def organization(self):
        return self.qt.organizationName()

    @organization.setter
    def organization(self, value):
        self.qt.setOrganizationName(value)

    @property
    def domain(self):
        return self.qt.organizationDomain()

    @domain.setter
    def domain(self, value):
        self.qt.setOrganizationDomain(value)

    def run(self):
        # Register interupt signal handler
        def signal_handler(signum, frame):
            if signum == signal.SIGINT:
                self.quit()
        signal.signal(signal.SIGINT, signal_handler)

        # Run timer to process interrupt signals
        timer = QtCore.QTimer()
        timer.timeout.connect(lambda: None)
        timer.start(250)

        return self.qt.exec_()

    def quit(self):
        self.qt.quit()

    @property
    def name_changed(self):
        return self.__name_changed

    @name_changed.setter
    def name_changed(self, value):
        self.__name_changed = value

    def __handle_name_changed(self):
        if callable(self.name_changed):
            self.name_changed(self.name)

    @property
    def version_changed(self):
        return self.__version_changed

    @version_changed.setter
    def version_changed(self, value):
        self.__version_changed = value

    def __handle_version_changed(self):
        if callable(self.version_changed):
            self.version_changed(self.version)

    @property
    def organization_changed(self):
        return self.__organization_changed

    @organization_changed.setter
    def organization_changed(self, value):
        self.__organization_changed = value

    def __handle_organization_changed(self):
        if callable(self.organization_changed):
            self.organization_changed(self.organization)

    @property
    def domain_changed(self):
        return self.__domain_changed

    @domain_changed.setter
    def domain_changed(self, value):
        self.__domain_changed = value

    def __handle_domain_changed(self):
        if callable(self.domain_changed):
            self.domain_changed(self.domain)

class GuiApplication(CoreApplication):

    QtClass = QtGui.QGuiApplication

    def __init__(self, name=None, *, display_name=None, icon=None,
                 display_name_changed=None, last_window_closed=None, **kwargs):
        super().__init__(name=name, **kwargs)
        if display_name is not None:
            self.display_name = display_name
        if icon is not None:
            self.icon = icon
        self.display_name_changed = display_name_changed
        self.last_window_closed = last_window_closed
        # Connect signals
        self.qt.applicationDisplayNameChanged.connect(self.__handle_display_name_changed)
        self.qt.lastWindowClosed.connect(self.__handle_last_window_closed)

    @property
    def display_name(self):
        return self.qt.applicationDisplayName()

    @display_name.setter
    def display_name(self, value):
        self.qt.setApplicationDisplayName(value)

    @property
    def icon(self):
        icon = self.qt.windowIcon()
        if icon.isNull():
            return None
        return Icon(qt=icon)

    @icon.setter
    def icon(self, value):
        if value is None:
            self.qt.setWindowIcon(QtGui.QIcon())
        else:
            if not isinstance(value, Icon):
                value = Icon(value)
            self.qt.setWindowIcon(value.qt)

    @property
    def display_name_changed(self):
        return self.__display_name_changed

    @display_name_changed.setter
    def display_name_changed(self, value):
        self.__display_name_changed = value

    def __handle_display_name_changed(self):
        if callable(self.display_name_changed):
            self.display_name_changed(self.display_name)

    @property
    def last_window_closed(self):
        return self.__last_window_closed

    @last_window_closed.setter
    def last_window_closed(self, value):
        self.__last_window_closed = value

    def __handle_last_window_closed(self):
        if callable(self.last_window_closed):
            self.last_window_closed()

class Application(GuiApplication):

    QtClass = QtWidgets.QApplication

    def __init__(self, name=None, *, focus_changed=None, **kwargs):
        super().__init__(name=name, **kwargs)
        self.focus_changed = focus_changed
        # Connect signals
        self.qt.focusChanged.connect(self.__handle_focus_changed)

    @property
    def focus_changed(self):
        return self.__focus_changed

    @focus_changed.setter
    def focus_changed(self, value):
        self.__focus_changed = value

    def __handle_focus_changed(self, old, now):
        if callable(self.focus_changed):
            self.focus_changed(old, now)

    def quit(self):
        """Request quit application by closing all windows."""
        self.qt.closeAllWindows()



"""Base classes of all user interface objects.

For more information on the underlying Qt5 object see [QWidget](https://doc.qt.io/qt-5/qwidget.html).
"""

from .qutie import QtCore
from .qutie import QtGui
from .qutie import QtWidgets

from .icon import Icon
from .object import Object

__all__ = [
    'Widget', 'BaseWidget',
]

class BaseWidget(Object):
    """Base widget for components without layout."""

    QtClass = QtWidgets.QWidget

    def __init__(self, *, title=None, width=None, height=None, enabled=None,
                 visible=None, status_tip=None, stylesheet=None, icon=None,
                 tool_tip=None, tool_tip_duration=None, parent=None,
                 close_event=None, focus_in=None, focus_out=None, **kwargs):
        super().__init__(**kwargs)
        if title is not None:
            self.title = title
        if width is not None:
            self.width = width
        if height is not None:
            self.height = height
        if visible is not None:
            self.visible = visible
        if enabled is not None:
            self.enabled = enabled
        if status_tip is not None:
            self.status_tip = status_tip
        if stylesheet is not None:
            self.stylesheet = stylesheet
        if icon is not None:
            self.icon = icon
        if tool_tip is not None:
            self.tool_tip = tool_tip
        if tool_tip_duration is not None:
            self.tool_tip_duration = tool_tip_duration
        if parent is not None:
            self.parent = parent
        self.close_event = close_event
        # Connect signals
        self.qt.closeEvent = self.__handle_close_event
        # Custom events
        self.focus_in = focus_in
        self.qt._focusInEvent = self.qt.focusInEvent
        def handle_focus_is(event):
            self.qt._focusInEvent(event)
            self.emit(self.focus_in)
        self.qt.focusInEvent = handle_focus_is
        self.focus_out = focus_out
        self.qt._focusOutEvent = self.qt.focusOutEvent
        def handle_focus_out(event):
            self.qt._focusOutEvent(event)
            self.emit(self.focus_out)
        self.qt.focusOutEvent = handle_focus_out

    @property
    def title(self):
        return self.qt.windowTitle()

    @title.setter
    def title(self, value):
        self.qt.setWindowTitle(value)

    @property
    def width(self):
        return self.qt.width()

    @property
    def minimum_width(self):
        return self.qt.minimumWidth()

    @minimum_width.setter
    def minimum_width(self, value):
        self.qt.setMinimumWidth(value)

    @property
    def maximum_width(self):
        return self.qt.maximumWidth()

    @maximum_width.setter
    def maximum_width(self, value):
        self.qt.setMaximumWidth(value)

    @width.setter
    def width(self, value):
        if value is None:
            self.qt.setMinimumWidth(0)
            self.qt.setMaximumWidth(QtWidgets.QWIDGETSIZE_MAX)
        else:
            self.qt.setMinimumWidth(value)
            self.qt.setMaximumWidth(value)

    @property
    def height(self):
        return self.qt.height()

    @property
    def minimum_height(self):
        return self.qt.minimumHeight()

    @minimum_height.setter
    def minimum_height(self, value):
        self.qt.setMinimumHeight(value)

    @property
    def maximum_height(self):
        return self.qt.maximumHeight()

    @maximum_height.setter
    def maximum_height(self, value):
        self.qt.setMaximumHeight(value)

    @height.setter
    def height(self, value):
        if value is None:
            self.qt.setMinimumHeight(0)
            self.qt.setMaximumHeight(QtWidgets.QWIDGETSIZE_MAX)
        else:
            self.qt.setMinimumHeight(value)
            self.qt.setMaximumHeight(value)

    @property
    def size(self):
        return self.width, self.height

    @size.setter
    def size(self, value):
        self.width = value[0]
        self.height = value[1]

    @property
    def minimum_size(self):
        return self.qt.minimumWidth(), self.qt.minimumHeight()

    @minimum_size.setter
    def minimum_size(self, value):
        return self.qt.setMinimumSize(value[0], value[1])

    @property
    def maximum_size(self):
        return self.qt.maximumWidth(), self.qt.maximumHeight()

    @maximum_size.setter
    def maximum_size(self, value):
        return self.qt.setMaximumSize(value[0], value[1])

    @property
    def position(self):
        return self.qt.x(), self.qt.y()

    @property
    def minimized(self):
        return self.qt.isMinimized()

    @minimized.setter
    def minimized(self, value):
        if value:
            self.qt.showMinimized()
        else:
            self.qt.showNormal()

    @property
    def maximized(self):
        return self.qt.isMaximized()

    @maximized.setter
    def maximized(self, value):
        if value:
            self.qt.showMaximized()
        else:
            self.qt.showNormal()

    @property
    def enabled(self):
        return self.qt.isEnabled()

    @enabled.setter
    def enabled(self, value):
        self.qt.setEnabled(value)

    @property
    def visible(self):
        return self.qt.isVisible()

    @visible.setter
    def visible(self, value):
        self.qt.setVisible(value)

    @property
    def status_tip(self):
        return self.qt.statusTip()

    @status_tip.setter
    def status_tip(self, value):
        self.qt.setStatusTip(value)

    @property
    def stylesheet(self):
        return self.qt.styleSheet()

    @stylesheet.setter
    def stylesheet(self, value):
        self.qt.setStyleSheet(value)

    @property
    def icon(self):
        icon = self.qt.windowIcon()
        if icon.isNull():
            return None
        return Icon(qt=icon)

    @icon.setter
    def icon(self, value):
        if value is None:
            self.qt.setWindowIcon(QtGui.QIcon())
        else:
            if not isinstance(value, Icon):
                value = Icon(value)
            self.qt.setWindowIcon(value.qt)

    @property
    def tool_tip(self):
        return self.qt.toolTip()

    @tool_tip.setter
    def tool_tip(self, value):
        self.qt.setToolTip(value)

    @property
    def tool_tip_duration(self):
        """Tooltip duration in seconds. Minimum duration is 1 millisecond."""
        return self.qt.toolTipDuration() / 1000.

    @tool_tip_duration.setter
    def tool_tip_duration(self, value):
        self.qt.setToolTipDuration(value * 1000.)

    @property
    def parent(self):
        parent = self.qt.parent()
        if hasattr(parent, 'reflection'):
            return parent.reflection()

    @parent.setter
    def parent(self, value):
        assert isinstance(value, BaseWidget), "Parent must inherit from BaseWidget"
        self.qt.setParent(value.qt)

    def close(self):
        """Close widget, return True if widget was closed."""
        return self.qt.close()

    def show(self):
        """Show widget and all child widgets."""
        self.qt.show()

    def hide(self):
        """Hide widget and all child widgets."""
        self.qt.hide()

    def down(self):
        """Lower widget to bottom of parent widget's stack."""
        self.qt.lower()

    def up(self):
        """Raise widget to top of parent widget's stack."""
        self.qt.raise_()

    def resize(self, width, height):
        self.qt.resize(width, height)

    def move(self, x, y):
        self.qt.move(x, y)

    def __handle_close_event(self, event):
        # Overwrite slot closeEvent
        if callable(self.close_event):
            if not self.close_event():
                event.ignore()
                return
        super(self.qt.__class__, self.qt).closeEvent(event)

class Widget(BaseWidget):
    """Widget for components with layout."""

    QtLayoutClass = QtWidgets.QVBoxLayout

    def __init__(self, *, layout=None, modal=None, **kwargs):
        super().__init__(**kwargs)
        self.qt.setLayout(self.QtLayoutClass())
        if layout is not None:
            self.layout = layout
        if modal is not None:
            self.modal = modal

    @property
    def layout(self):
        layout = self.qt.layout()
        if layout is not None:
            if layout.count():
                return layout.itemAt(0).widget().reflection()
        return None

    @layout.setter
    def layout(self, value):
        while self.qt.layout().count():
            self.qt.layout().takeAt(0)
        if value is not None:
            if not isinstance(value, BaseWidget):
                raise ValueError(value)
            self.qt.layout().addWidget(value.qt)

    @property
    def modal(self):
        return self.qt.isModal()

    @modal.setter
    def modal(self, value):
        modality = QtCore.Qt.ApplicationModal if value else QtCore.Qt.NonModal
        self.qt.setWindowModality(modality)



"""Single line text input.

For more information on the underlying Qt5 object see [QLineEdit](https://doc.qt.io/qt-5/qlineedit.html).
"""

from .qutie import QtWidgets

from .widget import BaseWidget

__all__ = ['Text']

class Text(BaseWidget):
    """Text input widget.

    >>> text = Text('lorem')
    >>> text.append(' ispum')
    >>> text.value
    'lorem ipsum'
    >>> tree.clear()
    text.value
    ''
    """

    QtClass = QtWidgets.QLineEdit

    def __init__(self, value=None, *, readonly=False, clearable=False,
                 changed=None, editing_finished=None, **kwargs):
        super().__init__(**kwargs)
        self.readonly = readonly
        self.clearable = clearable
        if value is not None:
            self.value = value
        self.changed = changed
        self.editing_finished = editing_finished
        # Connect signals
        self.qt.textChanged.connect(self.__handle_changed)
        self.qt.editingFinished.connect(self.__handle_editing_finished)

    @property
    def value(self):
        return self.qt.text()

    @value.setter
    def value(self, value):
        self.qt.setText(format(value))

    @property
    def readonly(self):
        """Read only enabled.

        >>> text.readonly = True
        >>> text.readonly
        True
        """
        return self.qt.isReadOnly()

    @readonly.setter
    def readonly(self, value):
        self.qt.setReadOnly(value)

    @property
    def clearable(self):
        """Inline clear button enabled.

        >>> text.clearable = True
        >>> text.clearable
        True
        """
        return self.qt.isClearButtonEnabled()

    @clearable.setter
    def clearable(self, value):
        self.qt.setClearButtonEnabled(bool(value))

    @property
    def changed(self):
        """Event called when value changed."""
        return self.__changed

    @changed.setter
    def changed(self, value):
        self.__changed = value

    def __handle_changed(self, text):
        if callable(self.changed):
            self.changed(text)

    @property
    def editing_finished(self):
        """Event called when editing finished."""
        return self.__editing_finished

    @editing_finished.setter
    def editing_finished(self, value):
        self.__editing_finished = value

    def __handle_editing_finished(self):
        if callable(self.editing_finished):
            self.editing_finished()

    def append(self, text):
        """Append text to current value.

        >>> text =Text('lorem')
        >>> text.append(' ipsum')
        >>> text.value
        'lorem ipsum'
        """
        self.value = ''.join((self.value, format(text)))

    def clear(self):
        """Clear current value.

        >>> text.clear()
        >>> text.value
        ''
        """
        self.qt.clear()



from .qutie import QtGui
from .qutie import QutieStub

from .pixmap import Pixmap

__all__ = ['Icon']

class Icon(QutieStub):
    """Icon containing multiple resolution pixmaps.

    Load multiple resoultion icons.
    >>> icon = Icon('small.png', 'large.png')

    Load icon from theme (X11 only).
    >>> icon = Icon.from_theme('document-open', Icon('fallback.png'))

    Create a color icon.
    >>> icon = Icon.from_color('red')

    Create a color icon of custom size.
    >>> icon = Icon.from_color('green', 32, 32)
    """

    QtClass = QtGui.QIcon

    def __init__(self, *items, **kwargs):
        super().__init__(**kwargs)
        for item in items:
            self.append(item)

    @property
    def available_sizes(self):
        sizes = []
        for size in self.qt.availableSizes():
            sizes.append((size.width(), size.height()))
        return tuple(sizes)

    @property
    def name(self):
        return self.qt.name()

    @property
    def theme_name(self):
        return self.QtClass.themeName()

    def append(self, value):
        if isinstance(value, str):
            # Try to create color from string
            color = QtGui.QColor(value)
            if color.isValid():
                pixmap = Pixmap.QtClass(64, 64)
                pixmap.fill(color)
                self.qt.addPixmap(pixmap)
            else:
                self.qt.addFile(value)
        elif isinstance(value, Pixmap):
            self.qt.addPixmap(value.qt)
        else:
            raise ValueError(value)

    def pixmap(self, width, height):
        return Pixmap(qt=self.qt.pixmap(width, height))

    @classmethod
    def from_color(cls, color, width=16, height=16):
        color = QtGui.QColor(color)
        pixmap = Pixmap.QtClass(width, height)
        pixmap.fill(color)
        return Icon(qt=pixmap)

    @classmethod
    def from_theme(cls, name, fallback=None):
        args = [name]
        if fallback is not None:
            args.append(fallback)
        icon = cls.QtClass.fromTheme(*args)
        return Icon(qt=icon)



"""A convenient worker thread class.

## Example Usage

```python
def calculate(worker):
    for i in range(100):
        # Emit custom events.
        worker.emit('progress', i, 100)
        worker.emit('message', "All ok...")

worker = ui.Worker(target=calculate)
# Assign custom event callbacks.
worker.progress = lambda step, max: print(f"progress: {step}/{max}")
worker.message = lambda msg: print(f"message: {msg}")
worker.start()
```

For more information on the underlying Qt5 object see [QObject](https://doc.qt.io/qt-5/qobject.html).
"""

import copy
import logging
import threading
import traceback

from .qutie import QtCore

from .object import Object

__all__ = ['Worker', 'StopRequest']

class StopRequest(Exception):
    """Raise to stop worker execution."""

class Worker(Object):
    """Worker thread class using python `threading`."""

    def __init__(self, *, target=None, started=None, finished=None, failed=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.target = target
        self.started = started
        self.finished = finished
        self.failed = failed
        self.__lock = threading.RLock()
        self.__thread = None
        self.__stop_requested = False
        self.__values = {}

    @property
    def target(self):
        return self.__target

    @target.setter
    def target(self, value):
        self.__target = value

    @property
    def started(self):
        return self.__started

    @started.setter
    def started(self, value):
        self.__started = value

    @property
    def finished(self):
        return self.__finished

    @finished.setter
    def finished(self, value):
        self.__finished = value

    @property
    def failed(self):
        """Event triggered if exception is thrown from worker thread. Provides
        arguments exception _e_ and traceback _tb_.

        >>> worker.failed = lambda e, tb: ...
        """
        return self.__failed

    @failed.setter
    def failed(self, value):
        self.__failed = value

    def set(self, key, value):
        """Set thread safe copy of value."""
        self.__set(key, value)

    def __set(self, key, value):
        with self.__lock:
            self.__values[key] = copy.deepcopy(value)

    def get(self, key, default=None):
        """Return thread safe copy of value."""
        return self.__get(key, default)

    def __get(self, key, default=None):
        with self.__lock:
            return copy.deepcopy(self.__values.get(key, default))

    def keys(self):
        """Return keys of available thread safe values."""
        return self.__keys()

    def __keys(self):
        with self.__lock:
            return tuple(self.__values.keys())

    def start(self):
        """Start worker thread."""
        self.__start()

    def __start(self):
        with self.__lock:
            if not self.__thread:
                self.__thread = threading.Thread(target=self.__run)
            self.__stop_requested = False
            self.__thread.start()

    def stop(self):
        """Stop worker. This will only work when using properties `stopping`
        or `running` in `target`.
        """
        self.__stop()

    def __stop(self):
        self.__stop_requested = True

    def join(self):
        """Join thread, blocking until finished."""
        self.__join()

    def __join(self):
        try:
            self.__thread.join()
        except AttributeError:
            pass

    @property
    def stopping(self):
        """Return `True` when stopping."""
        return self.__stop_requested

    @property
    def running(self):
        """Return `True` while not stopping."""
        return not self.__stop_requested

    @property
    def alive(self):
        """Return `True` while worker thread is alive."""
        with self.__lock:
            if self.__thread:
                return self.__thread.is_alive()
        return False

    def __run(self):
        self.emit('started')
        try:
            if self.target is None:
                self.run()
            else:
                self.target(self)
        except StopRequest:
            pass
        except Exception as e:
            tb = traceback.format_exc()
            logging.error("%s %s: %s", self, type(e).__name__, e)
            self.emit('failed', e, tb)
        finally:
            with self.__lock:
                self.__stop_requested = False
                self.__thread = None
                self.emit('finished')

    def run(self):
        pass



from .qutie import QtWidgets

from .action import Action
from .widget import BaseWidget

__all__ = ['Menu']

class Menu(BaseWidget):

    QtClass = QtWidgets.QMenu

    def __init__(self, *items, text=None, **kwargs):
        super().__init__(**kwargs)
        for item in items:
            self.append(item)
        if text is not None:
            self.text = text

    @property
    def text(self):
        return self.qt.title()

    @text.setter
    def text(self, value):
        self.qt.setTitle(value)

    def append(self, item):
        if isinstance(item, str):
            item = Action(item)
        if isinstance(item, Action):
            self.qt.addAction(item.qt)
        elif isinstance(item, Menu):
            self.qt.addMenu(item.qt)
        else:
            raise ValueError(item)
        return item

    def insert(self, index, item):
        if index < 0:
            index = max(0, len(self) + index)
        if isinstance(item, str):
            item = Action(item)
        before = self[index].qt if len(self) else None
        if isinstance(item, Action):
            self.qt.insertAction(before, item.qt)
        elif isinstance(item, Menu):
            self.qt.insertMenu(before, item.qt)
        else:
            raise ValueError(item)
        return item

    def index(self, item):
        return self.qt.actions().index(item.qt)

    def _get_action_or_menu(self, action):
        if hasattr(action, 'reflection'):
            return action.reflection()
        return action.menu().reflection()

    def __getitem__(self, index):
        action = self.qt.actions()[index]
        return self._get_action_or_menu(action)

    def __iter__(self):
        return iter(self._get_action_or_menu(action) for action in self.qt.actions())

    def __len__(self):
        return len(self.qt.actions())



"""Stack of tabbed widgets.

## Example Usage

>>> tabs = Tabs(
    Tab(title="Spam"),
    Tab(title="Ham")
)
>>> tabs.append(Tab("Eggs")) # append tab "Eggs"
>>> tabs.remove(tabs[0]) # remove tab "Spam"
>>> tabs.insert(0, Tab(title="Spam")) # insert again at begin
>>> tabs.current = tabs[1] # show tab "Ham"
>>> for tab in tabs:
...     print(tab.title)

For more information on the underlying Qt5 object see [QTabWidget](https://doc.qt.io/qt-5/qtabwidget.html).
"""

from .qutie import QtWidgets
from .widget import BaseWidget, Widget

__all__ = [
    'Tab',
    'Tabs'
]

class Tab(Widget):
    """Tab item, property `title` displayed in tab bar."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def title(self):
        """Title displayed in tab bar."""
        return self.qt.windowTitle()

    @title.setter
    def title(self, value):
        self.qt.setWindowTitle(value)
        # Update tab title
        parent = self.qt.parent()
        if parent:
            if parent.parent():
                index = parent.parent().indexOf(self.qt)
                parent.parent().setTabText(index, value)

class Tabs(BaseWidget):
    """Tab widget providign a tab bar and item stack."""

    QtClass = QtWidgets.QTabWidget

    def __init__(self, *items, changed=None, **kwargs):
        super().__init__(**kwargs)
        for item in items:
            self.append(item)
        self.changed = changed
        # Connect signals
        self.qt.currentChanged.connect(self.__handle_changed)

    @property
    def items(self):
        """Return list of tab items."""
        return list(self)

    def append(self, tab):
        """Append tab item to tab widget.

        >>> tabs.append(Tab(title="Spam"))
        """
        self.qt.addTab(tab.qt, tab.title)

    def insert(self, index, tab):
        """Insert tab item at index to tab widget.

        >>> tabs.insert(0, Tab(title="Spam"))
        """
        if index < 0:
            index = max(0, len(self) + index)
        self.qt.insertTab(index, tab.qt, tab.title)

    def remove(self, tab):
        """Remove tab item from tab widget.

        >>> tabs.remove(tabs[0])
        """
        index = self.qt.indexOf(tab.qt)
        self.qt.removeTab(index)

    @property
    def current(self):
        """Current active tab item.

        >>> tabs.current = tabs[0]
        >>> tabs.current
        <Tab ...>
        """
        index = self.qt.currentIndex()
        return self[index]

    @current.setter
    def current(self, item):
        index = self.index(item)
        self.qt.setCurrentIndex(index)

    def index(self, item):
        """Return index of tab item. Raising a `ValueError` if provided item is
        not part of the tab widget.

        >>> tabs.index(tabs[0])
        0
        """
        index = self.qt.indexOf(item.qt)
        if index < 0:
            raise ValueError("item not in tabs")
        return index

    def clear(self):
        """Remove all tab items."""
        while 0 != len(self):
            self.remove(self.current)

    @property
    def changed(self):
        """Event triggered if current tab item changed, either by clicking on a
        tab or setting property `current`.
        """
        return self.__changed

    @changed.setter
    def changed(self, changed):
        self.__changed = changed

    def __handle_changed(self, index):
        if callable(self.changed):
            self.changed(index)

    def __getitem__(self, key):
        widget = self.qt.widget(key)
        if not widget:
            raise KeyError(key)
        return widget.reflection()

    def __setitem__(self, key, value):
        del self[key]
        self.insert(key, value)

    def __delitem__(self, key):
        self.qt.removeTab(key)

    def __len__(self):
        return self.qt.count()

    def __iter__(self):
        return (self[index] for index in range(len(self)))



from .qutie import QtCore
from .qutie import Qutie

__all__ = ['Object']

class Object(Qutie):

    QtClass = QtCore.QObject

    def __init__(self, *, object_name=None, parent=None, destroyed=None,
                 object_name_changed=None, **kwargs):
        super().__init__(**kwargs)
        self.qt.setReflection(self)
        # Properties
        if object_name is not None:
            self.object_name = object_name
        if parent is not None:
            self.parent = parent
        self.destroyed = destroyed
        self.object_name_changed = object_name_changed
        # Connect signals
        self.qt.destroyed.connect(self.__handle_destroyed)
        self.qt.objectNameChanged.connect(self.__handle_object_name_changed)
        self.qt.handleEvent.connect(lambda event: event())

    @property
    def object_name(self) -> str:
        return self.qt.objectName()

    @object_name.setter
    def object_name(self, value: str):
        self.qt.setObjectName(value)

    @property
    def parent(self):
        parent = self.qt.parent()
        if hasattr(parent, 'reflection'):
            return parent.reflection()

    @parent.setter
    def parent(self, value):
        assert isinstance(value, Object), "Parent must inherit from Object"
        self.qt.setParent(value.qt)

    @property
    def destroyed(self) -> object:
        return self.__destroyed

    @destroyed.setter
    def destroyed(self, value: object):
        self.__destroyed = value

    def __handle_destroyed(self, obj: object):
        if callable(self.destroyed):
            self.destroyed()

    @property
    def object_name_changed(self) -> object:
        return self.__object_name_changed

    @object_name_changed.setter
    def object_name_changed(self, value: object):
        self.__object_name_changed = value

    def __handle_object_name_changed(self):
        if callable(self.object_name_changed):
            self.object_name_changed(self.object_name)

    def emit(self, *args, **kwargs):
        """Emit an event.

        >>> o.event = lambda value: print(value) # assign event callback
        >>> o.emit('event', 42)
        """
        if not args:
            raise ValueError("Missing event argument.")
        event = args[0]
        if isinstance(event, str):
            if hasattr(self, event):
                event = getattr(self, event)
        if callable(event):
            self.qt.handleEvent.emit(lambda: event(*args[1:], **kwargs))



import weakref

from .qutie import QtCore
from .qutie import QtWidgets

from .action import Action
from .menu import Menu
from .menubar import MenuBar
from .toolbar import ToolBar
from .widget import BaseWidget

__all__ = ['MainWindow']

class StatusBar(BaseWidget):

    QtClass = QtWidgets.QStatusBar

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def append(self, widget):
        self.qt.addPermanentWidget(widget.qt)
        return widget

class MainWindow(BaseWidget):

    QtClass = QtWidgets.QMainWindow

    class ToolBars:

        def __init__(self, qt):
            self.__qt = qt
            self.__toolbars = set()

        def clear(self):
            for item in self.__toolbars:
                self.__qt.removeToolBar(item.qt)
            self.__toolbars.clear()

        def add(self, item):
            if isinstance(item, str):
                item = ToolBar(title=item)
            self.__qt.addToolBar(item.qt)
            self.__toolbars.add(item)
            return item

        def remove(self, item):
            if item in self.__toolbars:
                self.__qt.removeToolBar(item.qt)
                self.__toolbars.remove(item)

        def __iter__(self):
            return iter(self.__toolbars)

        def __len__(self):
            return len(self.__toolbars)

    def __init__(self, *, layout=None, **kwargs):
        super().__init__(**kwargs)
        self.qt.setMenuBar(MenuBar().qt)
        self.qt.setStatusBar(StatusBar().qt)
        self.layout = layout
        self.__toolbars = self.ToolBars(self.qt)

    @property
    def layout(self):
        widget = self.qt.centralWidget()
        if widget is not None:
            return widget.reflection()
        return None

    @layout.setter
    def layout(self, value):
        if value is None:
            self.qt.setCentralWidget(None)
        else:
            if not isinstance(value, BaseWidget):
                raise ValueError(value)
            self.qt.setCentralWidget(value.qt)

    @property
    def menubar(self):
        return self.qt.menuBar().reflection()

    @property
    def statusbar(self):
        return self.qt.statusBar().reflection()

    @property
    def toolbars(self):
        return self.__toolbars



"""Storing persistent application settings on any platform using a context
manager instance.
"""

import os
import json

from appdirs import user_config_dir

from .application import Application

__all__ = ['Settings']

class Settings:
    """Storing persistent application settings on any platform.

    >>> with Settings() as settings:
    ...    settings['app'] = {'name': 'My App', users=['Monty', 'John'])
    ...    name = settings.get('app').get('name')
    """

    suffix = 'qutie'

    def __init__(self, persistent=True):

        self.persistent = persistent
        self.__settings = {}

    @property
    def persistent(self):
        return self.__persistent

    @persistent.setter
    def persistent(self, value):
        self.__persistent = value

    @property
    def filename(self):
        app = Application.instance()
        appname = app.name
        appauthor = app.organization
        path = user_config_dir(appname, appauthor)
        return f'{path}.{self.suffix}'

    def __enter__(self):
        """Read application settings from filesystem (if existing)."""
        filename = self.filename
        if os.path.isfile(filename):
            with open(filename, 'r') as f:
                try:
                    self.__settings = json.load(f)
                except json.JSONDecodeError:
                    self.__settings = {}
        return self.__settings

    def __exit__(self, *exc):
        """Write application settings to filesystem."""
        if self.persistent:
            filename = self.filename
            path = os.path.dirname(filename)
            if not os.path.exists(path):
                os.makedirs(path)
            with open(filename, 'w') as f:
                # Can raise exception!
                f.write(json.dumps(self.__settings))
        self.__settings = {}
        return False



"""Simple item based list view.

For more information on the underlying Qt5 objects see [QListWidget](https://doc.qt.io/qt-5/qlistwidget.html) and [QListWidgetItem](https://doc.qt.io/qt-5/qlistwidgetitem.html).
"""

from .qutie import QtCore
from .qutie import QtGui
from .qutie import QtWidgets
from .qutie import Qutie, QutieStub

from .icon import Icon
from .widget import BaseWidget

__all__ = ['List', 'ListItem']

class BaseItemView(BaseWidget):

    QtClass = QtWidgets.QAbstractItemView

    def __init__(self, *, icon_size=None, **kwargs):
        super().__init__(**kwargs)
        if icon_size is not None:
            self.icon_size = icon_size

    @property
    def icon_size(self):
        size = self.qt.iconSize()
        return size.width(), size.height()

    @icon_size.setter
    def icon_size(self, value):
        if isinstance(value, int):
            value = value, value
        self.qt.setIconSize(QtCore.QSize(*value))

class List(BaseItemView):

    QtClass = QtWidgets.QListWidget

    def __init__(self, items=None, *, view_mode=None, resize_mode=None,
                 changed=None, selected=None, clicked=None, double_clicked=None,
                 **kwargs):
        super().__init__(**kwargs)
        if items is not None:
            self.items = items
        if view_mode is not None:
            self.view_mode = view_mode
        if resize_mode is not None:
            self.resize_mode = resize_mode
        # HACK: adjust default icons size
        if 'icon_size' not in kwargs:
            self.icon_size = 16
        self.changed = changed
        self.selected = selected
        self.clicked = clicked
        self.double_clicked = double_clicked
        # Connect signals
        self.qt.currentItemChanged.connect(self.__handle_changed)
        self.qt.currentRowChanged[int].connect(self.__handle_selected)
        self.qt.itemClicked.connect(self.__handle_clicked)
        self.qt.itemDoubleClicked.connect(self.__handle_double_clicked)

    @property
    def items(self):
        return list(self)

    @items.setter
    def items(self, items):
        self.clear()
        for item in items:
            self.append(item)

    @property
    def view_mode(self):
        return {
            QtWidgets.QListView.ListMode: 'list',
            QtWidgets.QListView.IconMode: 'icon'
        }[self.qt.viewMode()]

    @view_mode.setter
    def view_mode(self, value):
        self.qt.setViewMode({
            'list': QtWidgets.QListView.ListMode,
            'icon': QtWidgets.QListView.IconMode
        }[value])

    @property
    def resize_mode(self):
        return {
            QtWidgets.QListView.Fixed: 'fixed',
            QtWidgets.QListView.Adjust: 'adjust'
        }[self.qt.resizeMode()]

    @resize_mode.setter
    def resize_mode(self, value):
        self.qt.setResizeMode({
            'fixed': QtWidgets.QListView.Fixed,
            'adjust': QtWidgets.QListView.Adjust
        }[value])

    @property
    def current(self):
        item = self.qt.currentItem()
        if item:
            return item.data(ListItem.QtPropertyRole)
        return None

    @current.setter
    def current(self, item):
        index = self.qt.row(item.qt)
        if index < 0:
            raise IndexError(item)
        self.qt.setCurrentItem(item.qt)

    def index(self, item):
        index = self.qt.row(item.qt)
        if index < 0:
            raise ValueError("item not in list")
        return index

    def clear(self):
        self.qt.clear()

    def append(self, item):
        if not isinstance(item, ListItem):
            item = ListItem(item)
        self.qt.addItem(item.qt)
        return item

    def insert(self, index, item):
        if index < 0:
            index = max(0, len(self) + index)
        if not isinstance(item, ListItem):
            item = ListItem(item)
        self.qt.insertItem(index, item.qt)
        return item

    def remove(self, item):
        if item is None:
            raise IndexError(item)
        index = self.qt.row(item.qt)
        if index < 0:
            raise IndexError(item)
        self.qt.takeItem(index)

    def scroll_to(self, item):
        self.qt.scrollToItem(item.qt)

    @property
    def changed(self):
        return self.__changed

    @changed.setter
    def changed(self, value):
        self.__changed = value

    def __handle_changed(self, current, _):
        if callable(self.changed):
            index = self.qt.row(current)
            item = self[index]
            self.changed(item.value, index)

    @property
    def selected(self):
        return self.__selected

    @selected.setter
    def selected(self, value):
        self.__selected = value

    def __handle_selected(self, index):
        if callable(self.selected):
            if index >= 0:
                value = self[index]
                self.selected(value, index)

    @property
    def clicked(self):
        return self.__clicked

    @clicked.setter
    def clicked(self, value):
        self.__clicked = value

    def __handle_clicked(self, item):
        if callable(self.clicked):
            data = item.data(ListItem.QtPropertyRole)
            if data is not None:
                index = self.qt.row(item)
                self.clicked(index, data)

    @property
    def double_clicked(self):
        return self.__double_clicked

    @double_clicked.setter
    def double_clicked(self, value):
        self.__double_clicked = value

    def __handle_double_clicked(self, item):
        if callable(self.double_clicked):
            data = item.data(ListItem.QtPropertyRole)
            if data is not None:
                index = self.qt.row(item)
                self.double_clicked(index, data)

    def __getitem__(self, key):
        item = self.qt.item(key)
        if not item:
            raise IndexError(key)
        return item.data(ListItem.QtPropertyRole)

    def __setitem__(self, key, value):
        item = ListItem(value)
        self.qt.takeItem(key)
        self.qt.insertItem(key, item.qt)

    def __delitem__(self, key):
        self.qt.takeItem(key)

    def __len__(self):
        return self.qt.count()

    def __iter__(self):
        return (self[row] for row in range(len(self)))

class ListItem(QutieStub):

    QtClass = QtWidgets.QListWidgetItem

    QtPropertyRole = 0x8000

    def __init__(self, value, *, color=None, background=None, icon=None,
                 enabled=True, checked=None, checkable=False, **kwargs):
        super().__init__(**kwargs)
        self.__default_foreground = self.qt.foreground()
        self.__default_background = self.qt.background()
        self.qt.setData(self.QtPropertyRole, self)
        self.value = value
        self.color = color
        self.background = background
        if icon is not None:
            self.icon = icon
        self.enabled = enabled
        self.checkable = checkable
        self.checked = checked

    @property
    def value(self):
        return self.qt.data(QtCore.Qt.UserRole)

    @value.setter
    def value(self, value):
        self.qt.setData(QtCore.Qt.UserRole, value)
        self.qt.setText(format(value))

    @property
    def color(self):
        return self.qt.foreground().color().name()

    @color.setter
    def color(self, value):
        if value is None:
            brush = QtGui.QBrush(self.__default_foreground)
        else:
            brush = QtGui.QBrush(QtGui.QColor(value))
        self.qt.setForeground(brush)

    @property
    def background(self):
        return self.qt.background().color().name()

    @background.setter
    def background(self, value):
        if value is None:
            brush = QtGui.QBrush(self.__default_background)
        else:
            brush = QtGui.QBrush(QtGui.QColor(value))
        self.qt.setBackground(brush)

    @property
    def icon(self):
        icon = self.qt.icon()
        if icon.isNull():
            return None
        return Icon(qt=icon)

    @icon.setter
    def icon(self, value):
        if value is None:
            self.qt.setIcon(QtGui.QIcon())
        else:
            if not isinstance(value, Icon):
                value = Icon(value)
            self.qt.setIcon(value.qt)

    @property
    def enabled(self):
        return bool(self.qt.flags() & QtCore.Qt.ItemIsEnabled)

    @enabled.setter
    def enabled(self, value):
        if value:
            self.qt.setFlags(self.qt.flags() | QtCore.Qt.ItemIsEnabled)
        else:
            self.qt.setFlags(self.qt.flags() & ~QtCore.Qt.ItemIsEnabled)

    @property
    def checked(self):
        return self.qt.checkState() == QtCore.Qt.Checked

    @checked.setter
    def checked(self, value):
        if value is None:
            flags = self.qt.flags() & ~QtCore.Qt.ItemIsUserCheckable
            self.qt.setFlags(flags)
        else:
            self.qt.setCheckState(QtCore.Qt.Checked if value else QtCore.Qt.Unchecked)

    @property
    def checkable(self):
        return self.qt.flags() & QtCore.Qt.ItemIsUserCheckable

    @checkable.setter
    def checkable(self, value):
        if value:
            flags = self.qt.flags() | QtCore.Qt.ItemIsUserCheckable
            self.qt.setCheckState(self.checked)
        else:
            flags = self.qt.flags() & ~QtCore.Qt.ItemIsUserCheckable
        self.qt.setFlags(flags)



# Provide backward compatibility for qutie<1.6.0

from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

__all__ = [
    'QtCore',
    'QtGui',
    'QtWidgets',
    'bind'
]

def bind(qt):
    """Decorator used to bind custom events on a Qt base class."""
    def bind(cls):
        cls.QtClass = qt
        return cls
    return bind



from .qutie import QtWidgets

from .widget import BaseWidget

__all__ = ['Number']

class Number(BaseWidget):

    QtClass = QtWidgets.QDoubleSpinBox

    prefix_format = "{} "
    suffix_format = " {}"

    def __init__(self, value=0, *, minimum=None, maximum=None, step=1.0,
                 decimals=0, prefix=None, suffix=None, readonly=False,
                 adaptive=False, special_value=None, changed=None,
                 editing_finished=None, **kwargs):
        super().__init__(**kwargs)
        self.minimum = -float('inf') if minimum is None else minimum
        self.maximum = float('inf') if maximum is None else maximum
        self.step = step
        self.decimals = decimals
        self.prefix = prefix
        self.suffix = suffix
        self.readonly = readonly
        self.adaptive = adaptive
        if special_value is not None:
            self.special_value = special_value
        self.value = value
        self.changed = changed
        self.editing_finished = editing_finished
        # Connect signals
        self.qt.valueChanged.connect(self.__handle_changed)
        self.qt.editingFinished.connect(self.__handle_editing_finished)

    @property
    def value(self):
        return self.qt.value()

    @value.setter
    def value(self, value):
        self.qt.setValue(value)

    @property
    def minimum(self):
        return self.qt.minimum()

    @minimum.setter
    def minimum(self, value):
        self.qt.setMinimum(value)

    @property
    def maximum(self):
        return self.qt.maximum()

    @maximum.setter
    def maximum(self, value):
        self.qt.setMaximum(value)

    @property
    def step(self):
        return self.qt.singleStep()

    @step.setter
    def step(self, value):
        self.qt.setSingleStep(value)

    @property
    def decimals(self):
        return self.qt.decimals()

    @decimals.setter
    def decimals(self, value):
        self.qt.setDecimals(value)

    @property
    def prefix(self):
        return self.qt.prefix().strip()

    @prefix.setter
    def prefix(self, prefix):
        if prefix is None:
            prefix = ""
        prefix = format(prefix).strip()
        if prefix:
            prefix = self.prefix_format.format(prefix)
        self.qt.setPrefix(prefix)

    @property
    def suffix(self):
        return self.qt.suffix().strip()

    @suffix.setter
    def suffix(self, suffix):
        if suffix is None:
            suffix = ""
        suffix = format(suffix).strip()
        if suffix:
            suffix = self.suffix_format.format(suffix)
        self.qt.setSuffix(suffix)

    @property
    def readonly(self):
        return self.qt.isReadOnly()

    @readonly.setter
    def readonly(self, value):
        self.qt.setReadOnly(value)
        if value:
            self.qt.setButtonSymbols(self.qt.NoButtons)
        else:
            self.qt.setButtonSymbols(self.qt.UpDownArrows)

    @property
    def adaptive(self):
        return self.qt.stepType() == self.qt.AdaptiveDecimalStepType

    @adaptive.setter
    def adaptive(self, value):
        if value:
            self.qt.setStepType(self.qt.AdaptiveDecimalStepType)
        else:
            self.qt.setStepType(self.qt.DefaultStepType)

    @property
    def special_value(self):
        return self.qt.specialValueText()

    @special_value.setter
    def special_value(self, value):
        self.qt.setSpecialValueText(value)

    @property
    def changed(self):
        return self.__changed

    @changed.setter
    def changed(self, value):
        self.__changed = value

    def __handle_changed(self, value):
        if callable(self.changed):
            self.changed(value)

    @property
    def editing_finished(self):
        return self.__editing_finished

    @editing_finished.setter
    def editing_finished(self, value):
        self.__editing_finished = value

    def __handle_editing_finished(self):
        if callable(self.editing_finished):
            self.editing_finished()



from .qutie import QtWidgets

from .action import Action
from .menu import Menu
from .widget import BaseWidget

__all__ = ['MenuBar']

class MenuBar(BaseWidget):

    QtClass = QtWidgets.QMenuBar

    def __init__(self, *items, **kwargs):
        super().__init__(**kwargs)
        for item in items:
            self.append(item)

    def index(self, item):
        for index, action in enumerate(self):
            if action is item:
                return index
        raise ValueError("item not in list")

    def clear(self):
        while len(self):
            self.remove(self[0])

    def append(self, item):
        if isinstance(item, str):
            item = Menu(text=item)
        if isinstance(item, Action):
            self.qt.addAction(item.qt)
        else:
            self.qt.addMenu(item.qt)
        return item

    def insert(self, index, item):
        if index < 0:
            index = max(0, len(self) + index)
        if isinstance(item, str):
            item = Menu(text=item)
        before = self[index] if len(self) else None
        if isinstance(before, type(None)):
            self.qt.insertMenu(None, item.qt)
        elif isinstance(before, Menu):
            self.qt.insertMenu(before.qt.menuAction(), item.qt)
        else:
            self.qt.insertMenu(before.qt, item.qt)
        return item

    def remove(self, item):
        index = self.index(item)
        self.qt.removeAction(self.qt.actions()[index])

    def _get_action_or_menu(self, action):
        if hasattr(action, 'reflection'):
            return action.reflection()
        return action.menu().reflection()

    def __getitem__(self, index):
        action = self.qt.actions()[index]
        return self._get_action_or_menu(action)

    def __iter__(self):
        return iter(self._get_action_or_menu(action) for action in self.qt.actions())

    def __len__(self):
        return len(self.qt.actions())



"""Simple item based table view.

For more information on the underlying Qt5 objects see [QTableWidget](https://doc.qt.io/qt-5/qtablewidget.html) and [QTableWidgetItem](https://doc.qt.io/qt-5/qtablewidgetitem.html).
"""

from .qutie import QtCore
from .qutie import QtGui
from .qutie import QtWidgets
from .qutie import Qutie, QutieStub

from .icon import Icon
from .list import BaseItemView

__all__ = ['Table', 'TableItem']

class Table(BaseItemView):
    """Table widget.

    >>> table = Table(header=["Key", "Value"])
    >>> table.append(["Spam", "Eggs"])
    >>> table.insert(0, ["Ham", "Spam"])
    >>> for row in table:
    ...     for item in row:
    ...         item.color = "blue"
    >>> table.clear()
    """

    QtClass = QtWidgets.QTableWidget

    def __init__(self, rows=None, *, header=None, stretch=True, sortable=False,
                 vertical_header=False, activated=None, changed=None,
                 clicked=None, double_clicked=None, selected=None, **kwargs):
        super().__init__(**kwargs)
        self.qt.horizontalHeader().setHighlightSections(False)
        self.qt.verticalHeader().setHighlightSections(False)
        self.header = header or []
        for row in rows or []:
            self.append(row)
        self.stretch = stretch
        self.sortable = sortable
        self.vertical_header = vertical_header
        self.selection_mode = 'single'
        self.selection_behavior = 'rows'
        self.activated = activated
        self.changed = changed
        self.clicked = clicked
        self.double_clicked = double_clicked
        self.selected = selected
        # Connect signals
        self.qt.itemActivated.connect(self.__handle_activated)
        self.qt.itemChanged.connect(self.__handle_changed)
        self.qt.itemClicked.connect(self.__handle_clicked)
        self.qt.itemDoubleClicked.connect(self.__handle_double_clicked)
        self.qt.itemSelectionChanged.connect(self.__handle_selected)

    @property
    def header(self):
        """Horizontal header labels.

        >>> table.header = 'Key', 'Value'
        >>> table.header
        ('Key', 'Value')
        """
        return tuple([self.qt.horizontalHeaderItem(index).text() for index in range(self.qt.columnCount())])

    @header.setter
    def header(self, value):
        self.qt.setColumnCount(len(value))
        self.qt.setHorizontalHeaderLabels(value)

    @property
    def vertical_header(self):
        return self.qt.verticalHeader().visible()

    @vertical_header.setter
    def vertical_header(self, value):
        self.qt.verticalHeader().setVisible(value)

    @property
    def sortable(self):
        return self.qt.isSortingEnabled()

    @sortable.setter
    def sortable(self, value):
        self.qt.setSortingEnabled(value)

    @property
    def selection_behavior(self):
        return {
            QtWidgets.QAbstractItemView.SelectItems: 'items',
            QtWidgets.QAbstractItemView.SelectRows: 'rows',
            QtWidgets.QAbstractItemView.SelectColumns: 'columns'
        }[self.qt.selectionBehavior()]

    @selection_behavior.setter
    def selection_behavior(self, value):
        self.qt.setSelectionBehavior({
            'items': QtWidgets.QAbstractItemView.SelectItems,
            'rows': QtWidgets.QAbstractItemView.SelectRows,
            'columns': QtWidgets.QAbstractItemView.SelectColumns
        }[value])

    @property
    def selection_mode(self):
        return {
            QtWidgets.QAbstractItemView.SingleSelection: 'single',
            QtWidgets.QAbstractItemView.ContiguousSelection: 'contiguous',
            QtWidgets.QAbstractItemView.ExtendedSelection: 'extended',
            QtWidgets.QAbstractItemView.MultiSelection: 'multi',
            QtWidgets.QAbstractItemView.NoSelection: 'no'
        }[self.qt.selectionMode()]

    @selection_mode.setter
    def selection_mode(self, value):
        self.qt.setSelectionMode({
            'single': QtWidgets.QAbstractItemView.SingleSelection,
            'contiguous': QtWidgets.QAbstractItemView.ContiguousSelection,
            'extended': QtWidgets.QAbstractItemView.ExtendedSelection,
            'multi': QtWidgets.QAbstractItemView.MultiSelection,
            'no': QtWidgets.QAbstractItemView.NoSelection
        }[value])

    @property
    def activated(self):
        return self.__activated

    @activated.setter
    def activated(self, value):
        self.__activated = value

    def __handle_activated(self, item):
        if callable(self.activated):
            data = item.data(item.UserType)
            if data is not None:
                self.activated(data)

    @property
    def changed(self):
        return self.__changed

    @changed.setter
    def changed(self, value):
        self.__changed = value

    def __handle_changed(self, item):
        if callable(self.changed):
            data = item.data(item.UserType)
            if data is not None:
                self.changed(data)

    @property
    def clicked(self):
        return self.__clicked

    @clicked.setter
    def clicked(self, value):
        self.__clicked = value

    def __handle_clicked(self, item):
        if callable(self.clicked):
            data = item.data(item.UserType)
            if data is not None:
                self.clicked(data)

    @property
    def double_clicked(self):
        return self.__double_clicked

    @double_clicked.setter
    def double_clicked(self, value):
        self.__double_clicked = value

    def __handle_double_clicked(self, item):
        if callable(self.double_clicked):
            data = item.data(item.UserType)
            if data is not None:
                self.double_clicked(data)

    @property
    def selected(self):
        return self.__selected

    @selected.setter
    def selected(self, value):
        self.__selected = value

    def __handle_selected(self):
        if callable(self.selected):
            data = self.qt.selectedItems()
            if data:
                first = data[0].data(data[0].UserType)
                self.selected(first)

    @property
    def row_count(self):
        """Return row count."""
        return self.qt.rowCount()

    @property
    def column_count(self):
        """Return column count."""
        return self.qt.columnCount()

    def row(self, item):
        """Return item row."""
        return self.qt.row(item.qt)

    def column(self, item):
        """Return item column."""
        return self.qt.column(item.qt)

    def append(self, items):
        """Append items, returns appended items.

        >>> table.append([TableItem(value="Spam"), TableItem(value="Eggs")])
        or
        >>> table.append(["Spam", "Eggs"])
        """
        row = self.row_count
        return self.insert(row, items)

    def insert(self, row, items):
        """Insert items at row, returns inserted items.

        >>> table.insert(0, [TableItem(value="Spam"), TableItem(value="Eggs")])
        or
        >>> table.insert(0, ["Spam", "Eggs"])
        """
        self.qt.insertRow(row)
        for column, item in enumerate(items):
            if not isinstance(item, TableItem):
                item = TableItem(value=item)
            self.qt.setItem(row, column, item.qt)
            self.qt.resizeRowToContents(row)
        return self[row]

    def remove_row(self, row):
        if row < 0:
            row = max(0, self.row_count + row)
        self.qt.removeRow(row)

    def remove_column(self, column):
        if column < 0:
            column = max(0, self.column_count + column)
        self.qt.removeColumn(column)

    def clear(self):
        """Clear table contents."""
        self.qt.clearContents()

    @property
    def current(self):
        """Returns current table item or None."""
        item = self.qt.currentItem()
        if item is not None:
            return item.data(item.UserType)
        return None

    @current.setter
    def current(self, value):
        assert isinstance(value, TableItem)
        self.qt.setCurrentItem(value.qt)

    @property
    def stretch(self):
        """Stretch last column.

        >>> table.stretch = True
        >>> table.stretch
        True
        """
        return self.qt.horizontalHeader().stretchLastSection()

    @stretch.setter
    def stretch(self, value):
        self.qt.horizontalHeader().setStretchLastSection(value)

    def fit(self, column=None):
        if column is None:
            self.qt.resizeColumnsToContents()
        else:
            self.qt.resizeColumnToContents(column)
        self.qt.resizeRowsToContents()

    def __getitem__(self, key):
        items = []
        for column in range(self.column_count):
            item = self.qt.item(key, column)
            if item:
                items.append(item.data(item.UserType))
            else:
                items.append(None)
        return items

    def __setitem__(self, key, value):
        self.qt.removeRow(key)
        self.qt.insertRow(key)
        for column, item in enumerate(value):
            if not isinstance(item, TableItem):
                item = TableItem(value=item)
            self.qt.setItem(key, column, item.qt)

    def __delitem__(self, key):
        self.remove_row(key)

    def __len__(self):
        return self.row_count

    def __iter__(self):
        return (self[row] for row in range(len(self)))

class TableItem(QutieStub):

    QtClass = QtWidgets.QTableWidgetItem

    def __init__(self, value=None, *, color=None, background=None, icon=None,
                 enabled=True, readonly=True, checked=None, checkable=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.__default_foreground = self.qt.foreground()
        self.__default_background = self.qt.background()
        self.qt.setData(self.qt.UserType, self)
        self.value = value
        self.color = color
        self.background = background
        if icon is not None:
            self.icon = icon
        self.enabled = enabled
        self.readonly = readonly
        if checkable is not None:
            self.checkable = checkable
        if checked is not None:
            self.checked = checked

    @property
    def value(self):
        return self.qt.data(self.qt.Type)

    @value.setter
    def value(self, value):
        return self.qt.setData(self.qt.Type, value)

    @property
    def color(self):
        return self.qt.foreground().color().name()

    @color.setter
    def color(self, value):
        if value is None:
            brush = QtGui.QBrush(self.__default_foreground)
        else:
            brush = QtGui.QBrush(QtGui.QColor(value))
        self.qt.setForeground(brush)

    @property
    def background(self):
        return self.qt.background().color().name()

    @background.setter
    def background(self, value):
        if value is None:
            brush = QtGui.QBrush(self.__default_background)
        else:
            brush = QtGui.QBrush(QtGui.QColor(value))
        self.qt.setBackground(brush)

    @property
    def icon(self):
        icon = self.qt.icon()
        if icon.isNull():
            return None
        return Icon(qt=icon)

    @icon.setter
    def icon(self, value):
        if value is None:
            self.qt.setIcon(QtGui.QIcon())
        else:
            if not isinstance(value, Icon):
                value = Icon(value)
            self.qt.setIcon(value.qt)

    @property
    def enabled(self):
        return bool(self.qt.flags() & QtCore.Qt.ItemIsEnabled)

    @enabled.setter
    def enabled(self, value):
        if value:
            self.qt.setFlags(self.qt.flags() | QtCore.Qt.ItemIsEnabled)
        else:
            self.qt.setFlags(self.qt.flags() & ~QtCore.Qt.ItemIsEnabled)

    @property
    def readonly(self):
        return bool(self.qt.flags() & QtCore.Qt.ItemIsEditable)

    @readonly.setter
    def readonly(self, value):
        if value:
            self.qt.setFlags(self.qt.flags() & ~QtCore.Qt.ItemIsEditable)
        else:
            self.qt.setFlags(self.qt.flags() | QtCore.Qt.ItemIsEditable)

    @property
    def checked(self):
        return self.qt.checkState() == QtCore.Qt.Checked

    @checked.setter
    def checked(self, value):
        if value is None:
            flags = self.qt.flags() & ~QtCore.Qt.ItemIsUserCheckable
            self.qt.setFlags(flags)
        else:
            self.qt.setCheckState(QtCore.Qt.Checked if value else QtCore.Qt.Unchecked)

    @property
    def checkable(self):
        return self.qt.flags() & QtCore.Qt.ItemIsUserCheckable

    @checkable.setter
    def checkable(self, value):
        if value:
            flags = self.qt.flags() | QtCore.Qt.ItemIsUserCheckable
        else:
            flags = self.qt.flags() & ~QtCore.Qt.ItemIsUserCheckable
        self.qt.setFlags(flags)
        self.qt.setCheckState(self.checked)



"""Simple item based tree view.

For more information on the underlying Qt5 objects see [QTreeWidget](https://doc.qt.io/qt-5/qtreewidget.html) and [QTreeWidgetItem](https://doc.qt.io/qt-5/qtreewidgetitem.html).
"""

from .qutie import QtCore
from .qutie import QtGui
from .qutie import QtWidgets
from .qutie import Qutie, QutieStub

from .icon import Icon
from .list import BaseItemView

__all__ = ['Tree', 'TreeItem', 'TreeItemColumn']

class Tree(BaseItemView):
    """Tree widget.

    >>> tree = Tree(header=["Key", "Value"])
    >>> tree.append(["Spam", "Eggs"])
    >>> for item in tree:
    ...     item[0].checked =True
    ...     item[1].color = "blue"
    ...     child = item.append(["Ham", "Spam"])
    ...     child.checked = False
    >>> tree.clear()
    """

    QtClass = QtWidgets.QTreeWidget

    def __init__(self, items=None, *, expands_on_double_click=None, header=None,
                 sortable=False, indentation=None, root_is_decorated=None,
                 uniform_row_heights=None, word_wrap=None, activated=None,
                 changed=None, clicked=None, double_clicked=None, selected=None,
                 **kwargs):
        super().__init__(**kwargs)
        if items is not None:
            self.items = items
        if expands_on_double_click is not None:
            self.expands_on_double_click = expands_on_double_click
        self.header = header or []
        self.sortable = sortable
        if indentation is not None:
            self.indentation = indentation
        if root_is_decorated is not None:
            self.root_is_decorated = root_is_decorated
        if uniform_row_heights is not None:
            self.uniform_row_heights = uniform_row_heights
        if word_wrap is not None:
            self.word_wrap = word_wrap
        self.activated = activated
        self.changed = changed
        self.clicked = clicked
        self.double_clicked = double_clicked
        self.selected = selected
        # Connect signals
        self.qt.itemActivated.connect(self.__handle_activated)
        self.qt.itemChanged.connect(self.__handle_changed)
        self.qt.itemClicked.connect(self.__handle_clicked)
        self.qt.itemDoubleClicked.connect(self.__handle_double_clicked)
        self.qt.itemSelectionChanged.connect(self.__handle_selected)

    @property
    def expands_on_double_click(self):
        return self.qt.expandsOnDoubleClick()

    @expands_on_double_click.setter
    def expands_on_double_click(self, value):
        self.qt.setExpandsOnDoubleClick(bool(value))

    @property
    def header(self):
        item = self.qt.headerItem()
        return tuple([item.text(index) for index in range(self.qt.columnCount())])

    @header.setter
    def header(self, value):
        self.qt.setColumnCount(len(value))
        self.qt.setHeaderLabels(value)

    @property
    def sortable(self):
        return self.qt.isSortingEnabled()

    @sortable.setter
    def sortable(self, value):
        self.qt.setSortingEnabled(value)

    @property
    def indentation(self):
        return self.qt.indentation()

    @indentation.setter
    def indentation(self, value):
        if value is None:
            self.qt.resetIndentation()
        else:
            self.qt.setIndentation(int(value))

    @property
    def root_is_decorated(self):
        return self.qt.rootIsDecorated()

    @root_is_decorated.setter
    def root_is_decorated(self, value):
        self.qt.setRootIsDecorated(bool(value))

    @property
    def uniform_row_heights(self):
        return self.qt.uniformRowHeights()

    @uniform_row_heights.setter
    def uniform_row_heights(self, value):
        self.qt.setUniformRowHeights(bool(value))

    @property
    def word_wrap(self):
        return self.qt.wordWrap()

    @word_wrap.setter
    def word_wrap(self, value):
        self.qt.setWordWrap(bool(value))

    @property
    def activated(self):
        return self.__activated

    @activated.setter
    def activated(self, value):
        self.__activated = value

    def __handle_activated(self, item, index):
        if callable(self.activated):
            data = item.data(0, item.UserType)
            if data is not None:
                self.activated(index, data)

    @property
    def changed(self):
        return self.__changed

    @changed.setter
    def changed(self, value):
        self.__changed = value

    def __handle_changed(self, item, index):
        if callable(self.changed):
            data = item.data(0, item.UserType)
            if data is not None:
                self.changed(index, data)

    @property
    def clicked(self):
        return self.__clicked

    @clicked.setter
    def clicked(self, value):
        self.__clicked = value

    def __handle_clicked(self, item, index):
        if callable(self.clicked):
            data = item.data(0, item.UserType)
            if data is not None:
                self.clicked(index, data)

    @property
    def double_clicked(self):
        return self.__double_clicked

    @double_clicked.setter
    def double_clicked(self, value):
        self.__double_clicked = value

    def __handle_double_clicked(self, item, index):
        if callable(self.double_clicked):
            data = item.data(0, item.UserType)
            if data is not None:
                self.double_clicked(index, data)

    @property
    def selected(self):
        return self.__selected

    @selected.setter
    def selected(self, value):
        self.__selected = value

    def __handle_selected(self):
        if callable(self.selected):
            items = self.qt.selectedItems()
            if items:
                first = items[0].data(0, items[0].UserType)
                self.selected(first)

    def append(self, item):
        """Append item or item labels, returns appended item.
        >>> tree.append(TreeItem(["Spam", "Eggs"]))
        or
        >>> tree.append(["Spam", "Eggs"])
        """
        if not isinstance(item, TreeItem):
            item = TreeItem(item)
        self.qt.addTopLevelItem(item.qt)
        item.expanded = True
        return item

    def insert(self, index, item):
        """Insert item or item labels at index, returns inserted item.
        >>> tree.insert(0, TreeItem(["Spam", "Eggs"]))
        or
        >>> tree.insert(0, ["Spam", "Eggs"])
        """
        if not isinstance(item, TreeItem):
            item = TreeItem(item)
        self.qt.insertTopLevelItem(index, item.qt)
        item.expanded = True
        return item

    def remove(self, item):
        index = self.qt.indexOfTopLevelItem(item.qt)
        self.qt.takeTopLevelItem(index)

    def clear(self):
        self.qt.clear()

    @property
    def current(self):
        """Returns current tree item or None."""
        item = self.qt.currentItem()
        if item is not None:
            return item.data(0, item.UserType)
        return item

    @current.setter
    def current(self, value):
        assert isinstance(value, TreeItem)
        self.qt.setCurrentItem(value.qt)

    def index(self, item):
        return self.qt.indexOfTopLevelItem(item.qt)

    @property
    def stretch(self):
        return self.qt.header().stretchLastSection()

    @stretch.setter
    def stretch(self, value):
        self.qt.header().setStretchLastSection(value)

    def fit(self, column=None):
        if column is None:
            for column in range(self.qt.columnCount()):
                self.qt.resizeColumnToContents(column)
        else:
            self.qt.resizeColumnToContents(column)

    def scroll_to(self, item):
        """Scroll to item to ensure item is visible."""
        self.qt.scrollToItem(item.qt)

    def __getitem__(self, key):
        item = self.qt.topLevelItem(key)
        return item.data(0, item.UserType)

    def __setitem__(self, key, value):
        self.remove(value)
        self.insert(key, value)

    def __delitem__(self, key):
        item = self.qt.topLevelItem(key)
        self.remove(item)

    def __len__(self):
        return self.qt.topLevelItemCount()

    def __iter__(self):
        return (self[index] for index in range(len(self)))

class TreeItem(QutieStub):
    """Tree item class."""

    QtClass = QtWidgets.QTreeWidgetItem

    def __init__(self, values=None, **kwargs):
        super().__init__(**kwargs)
        self.qt._default_foreground = self.qt.foreground(0)
        self.qt._default_background = self.qt.background(0)
        self.qt.setData(0, self.qt.UserType, self)
        if values is not None:
            for column, value in enumerate(values):
                self[column].value = value

    @property
    def children(self):
        """List of tree item's children."""
        items = []
        for index in range(self.qt.childCount()):
            item = self.qt.child(index)
            items.append(item.data(0, item.UserType))
        return items

    def append(self, item):
        """Append child item to this item."""
        if not isinstance(item, TreeItem):
            item = TreeItem(item)
        self.qt.addChild(item.qt)
        return item

    def insert(self, index, item):
        """Insert child item to this item."""
        if not isinstance(item, TreeItem):
            item = TreeItem(item)
        self.qt.insertChild(index, item.qt)
        return item

    @property
    def checkable(self):
        """Checkable state, `True` if item is checkable by user."""
        return self.qt.flags() & QtCore.Qt.ItemIsUserCheckable

    @checkable.setter
    def checkable(self, value):
        if value:
            flags = self.qt.flags() | QtCore.Qt.ItemIsUserCheckable
        else:
            flags = self.qt.flags() & ~QtCore.Qt.ItemIsUserCheckable
        self.qt.setFlags(flags)

    @property
    def expanded(self):
        """Expanded state, `True` if item is expanded."""
        return self.qt.isExpanded()

    @expanded.setter
    def expanded(self, value):
        self.qt.setExpanded(value)

    def __getitem__(self, column):
        return TreeItemColumn(column, self.qt)

    def __len__(self):
        return self.qt.columnCount()

    def __iter__(self):
        return (TreeItemColumn(column, self.qt) for column in range(len(self)))

class TreeItemColumn:
    """This class provides access to tree item column specific properties."""

    def __init__(self, column, qt):
        self.__column = column
        self.__qt = qt

    @property
    def column(self):
        return self.__column

    @property
    def qt(self):
        return self.__qt

    @property
    def value(self):
        """Column value."""
        return self.qt.data(self.column, self.qt.Type)

    @value.setter
    def value(self, value):
        return self.qt.setData(self.column, self.qt.Type, value)

    @property
    def color(self):
        return self.qt.foreground(self.column).color().name()

    @color.setter
    def color(self, value):
        if value is None:
            brush = QtGui.QBrush(self.qt._default_foreground)
        else:
            brush = QtGui.QBrush(QtGui.QColor(value))
        self.qt.setForeground(self.column, brush)

    @property
    def background(self):
        return self.qt.background(self.column).color().name()

    @background.setter
    def background(self, value):
        if value is None:
            brush = QtGui.QBrush(self.qt._default_background)
        else:
            brush = QtGui.QBrush(QtGui.QColor(value))
        self.qt.setBackground(self.column, brush)

    @property
    def icon(self):
        """Column icon, can be a `Pixmap`, filename or color."""
        icon = self.qt.icon(self.column)
        if icon.isNull():
            return None
        return Icon(qt=icon)

    @icon.setter
    def icon(self, value):
        if value is None:
            self.qt.setIcon(self.column, QtGui.QIcon())
        else:
            if not isinstance(value, Icon):
                value = Icon(value)
            self.qt.setIcon(self.column, value.qt)

    @property
    def checked(self):
        return self.qt.checkState(self.column) == QtCore.Qt.Checked

    @checked.setter
    def checked(self, value):
        self.qt.setCheckState(self.column, QtCore.Qt.Checked if value else QtCore.Qt.Unchecked)



"""Splitter widget.

For more information on the underlying Qt5 object see [QSplitter](https://doc.qt.io/qt-5/qsplitter.html).
"""

from .qutie import QtWidgets

from .widget import BaseWidget
from .mixins import OrientationMixin

__all__ = ['Splitter']

class Splitter(BaseWidget, OrientationMixin):
    """Splitter

    >>> splitter = Splitter(orientation='vertical')
    >>> splitter.append(List(["Spam", "Eggs"]))
    >>> splitter.insert(List(["Ham", "Spam"]))
    >>> for child in splitter.children:
    ...     pass
    """

    QtClass = QtWidgets.QSplitter

    def __init__(self, *children, sizes=None, orientation=None,
                 collapsible=True, stretch=None, **kwargs):
        super().__init__(**kwargs)
        for child in children:
            self.append(child)
        if sizes is not None:
            self.sizes = sizes
        if orientation is not None:
            self.orientation = orientation
        self.collapsible = collapsible
        if stretch is not None:
            for index, value in enumerate(stretch):
                self.qt.setStretchFactor(index, value)

    @property
    def children(self):
        children = []
        for index in range(self.qt.count()):
            children.append(self.qt.widget(index).reflection())
        return tuple(children)

    @property
    def sizes(self):
        return tuple(self.qt.sizes())

    @sizes.setter
    def sizes(self, value):
        self.qt.setSizes(list(value))

    @property
    def collapsible(self):
        return self.qt.childrenCollapsible()

    @collapsible.setter
    def collapsible(self, value):
        self.qt.setChildrenCollapsible(value)

    @property
    def handle_width(self):
        return self.qt.handleWidth()

    @handle_width.setter
    def handle_width(self, value):
        self.qt.setHandleWidth(value)

    def append(self, child):
        if not isinstance(child, BaseWidget):
            raise ValueError(child)
        self.qt.addWidget(child.qt)

    def __getitem__(self, index):
        item = self.qt.widget(index)
        if item is not None:
            return item.reflection()
        return None

    def __len__(self):
        return self.qt.count()

    def __iter__(self):
        return (self.qt.widget(index).reflection() for index in range(len(self)))



from .qutie import QtGui
from .qutie import QtWidgets

from .object import Object
from .icon import Icon

__all__ = ['Action']

class Action(Object):

    QtClass = QtWidgets.QAction

    def __init__(self, text=None, *, checkable=None, checked=False, tool_tip=None,
                 status_tip=None, shortcut=None, icon=None, separator=None,
                 triggered=None, toggled=None, **kwargs):
        super().__init__(**kwargs)
        # Create separator from text
        if isinstance(text, str) and text.startswith("---"):
            text = None
            separator = True
        if text is not None:
            self.text = text
        if checkable is not None:
            self.checkable = checkable
        self.checked = checked
        if tool_tip is not None:
            self.tool_tip = tool_tip
        if status_tip is not None:
            self.status_tip = status_tip
        if shortcut is not None:
            self.shortcut = shortcut
        if icon is not None:
            self.icon = icon
        if separator is not None:
            self.separator = separator
        self.triggered = triggered
        self.toggled = toggled
        # Connect signals
        self.qt.triggered.connect(self.__handle_triggered)
        self.qt.toggled.connect(self.__handle_toggled)

    @property
    def text(self):
        return self.qt.text()

    @text.setter
    def text(self, value):
        self.qt.setText(value)

    @property
    def checked(self):
        return self.qt.isChecked()

    @checked.setter
    def checked(self, value):
        self.qt.setChecked(value)

    @property
    def checkable(self):
        return self.qt.isCheckable()

    @checkable.setter
    def checkable(self, value):
        self.qt.setCheckable(value)

    @property
    def tool_tip(self):
        return self.qt.toolTip()

    @tool_tip.setter
    def tool_tip(self, value):
        self.qt.setToolTip(value)

    @property
    def status_tip(self):
        return self.qt.statusTip()

    @status_tip.setter
    def status_tip(self, value):
        self.qt.setStatusTip(value)

    @property
    def shortcut(self):
        return self.qt.shortcut().toString() or None

    @shortcut.setter
    def shortcut(self, value):
        if value is None:
            self.qt.setShortcut(QtGui.QKeySequence())
        else:
            self.qt.setShortcut(QtGui.QKeySequence(value))

    @property
    def icon(self):
        icon = self.qt.icon()
        if icon.isNull():
            return None
        return Icon(qt=icon)

    @icon.setter
    def icon(self, value):
        if value is None:
            self.qt.setIcon(QtGui.QIcon())
        else:
            if not isinstance(value, Icon):
                value = Icon(value)
            self.qt.setIcon(value.qt)

    @property
    def separator(self):
        return self.qt.isSeparator()

    @separator.setter
    def separator(self, value):
        self.qt.setSeparator(value)

    @property
    def triggered(self):
        return self.__triggered

    @triggered.setter
    def triggered(self, value):
        self.__triggered = value

    def __handle_triggered(self):
        if callable(self.triggered):
            self.triggered()

    @property
    def toggled(self):
        return self.__toggled

    @toggled.setter
    def toggled(self, value):
        self.__toggled = value

    def __handle_toggled(self, checked):
        if callable(self.toggled):
            self.toggled(checked)

    def trigger(self):
        self.qt.trigger()

    def toggle(self):
        self.qt.toggle()



"""A scrollable area.

For more information on the underlying Qt5 object see [QScrollArea](https://doc.qt.io/qt-5/qscrollarea.html).
"""

from .qutie import QtGui
from .qutie import QtWidgets

from .widget import Widget

__all__ = ['ScrollArea']

class ScrollArea(Widget):
    """A scrollable area."""

    QtClass = QtWidgets.QScrollArea

    def __init__(self, layout=None, **kwargs):
        super().__init__(**kwargs)
        self.qt.setWidgetResizable(True)
        self.qt.setWidget(Widget().qt)
        self.qt.setBackgroundRole(QtGui.QPalette.Base) # fix background
        if layout is not None:
            self.layout = layout

    @property
    def layout(self):
        return self.qt.widget().reflection().layout

    @layout.setter
    def layout(self, value):
        if self.qt.widget():
            self.qt.widget().reflection().layout = value



from .qutie import QtCore
from .qutie import QtWidgets

from .action import Action
from .menu import Menu
from .widget import BaseWidget

__all__ = ['ToolBar']

class ToolBar(BaseWidget):

    QtClass = QtWidgets.QToolBar

    def __init__(self, *items, floatable=None, movable=None, orientation=None,
                 tool_button_style=None, orientation_changed=None,
                 tool_button_style_changed=None, **kwargs):
        super().__init__(**kwargs)
        if floatable is not None:
            self.floatable = floatable
        if movable is not None:
            self.movable = movable
        if orientation is not None:
            self.orientation = orientation
        if tool_button_style is not None:
            self.tool_button_style = tool_button_style
        for item in items:
            self.append(item)
        self.orientation_changed = orientation_changed
        self.tool_button_style_changed = tool_button_style_changed
        # Connect signals
        self.qt.orientationChanged.connect(self.__handle_orientation_changed)

    @property
    def floatable(self):
        return self.qt.isFloatable()

    @floatable.setter
    def floatable(self, value):
        self.qt.setFloatable(bool(value))

    @property
    def movable(self):
        return self.qt.isMovable()

    @movable.setter
    def movable(self, value):
        self.qt.setMovable(bool(value))

    @property
    def orientation(self):
        return {
            QtCore.Qt.Horizontal: 'horizontal',
            QtCore.Qt.Vertical: 'vertical'
        }[self.qt.orientation()]

    @orientation.setter
    def orientation(self, value):
        self.qt.setOrientation({
            'horizontal': QtCore.Qt.Horizontal,
            'vertical': QtCore.Qt.Vertical
        }[value])

    @property
    def tool_button_style(self):
        return {
            QtCore.Qt.ToolButtonIconOnly: 'icon_only',
            QtCore.Qt.ToolButtonTextOnly: 'text_only',
            QtCore.Qt.ToolButtonTextBesideIcon: 'text_beside_icon',
            QtCore.Qt.ToolButtonTextUnderIcon: 'text_under_icon',
            QtCore.Qt.ToolButtonFollowStyle: 'follow_style'
        }[self.qt.toolButtonStyle()]

    @tool_button_style.setter
    def tool_button_style(self, value):
        self.qt.setToolButtonStyle({
            'icon_only': QtCore.Qt.ToolButtonIconOnly,
            'text_only': QtCore.Qt.ToolButtonTextOnly,
            'text_beside_icon': QtCore.Qt.ToolButtonTextBesideIcon,
            'text_under_icon': QtCore.Qt.ToolButtonTextUnderIcon,
            'follow_style': QtCore.Qt.ToolButtonFollowStyle
        }[value])

    def index(self, item):
        for index, action in enumerate(self):
            if action is item:
                return index
        raise ValueError("item not in list")

    def clear(self):
        while len(self):
            self.remove(self[0])

    def append(self, item):
        if isinstance(item, str):
            item = Action(text=item)
        if isinstance(item, Menu):
            self.qt.addAction(item.qt.menuAction())
        else:
            self.qt.addAction(item.qt)
        return item

    def insert(self, index, item):
        if index < 0:
            index = max(0, len(self) + index)
        if isinstance(item, str):
            item = Action(text=item)
        before = self[index] if len(self) else None
        if isinstance(before, type(None)):
            self.qt.insertAction(None, item.qt)
        elif isinstance(before, Menu):
            self.qt.insertAction(before.qt.menuAction(), item.qt)
        else:
            self.qt.insertAction(before.qt, item.qt)
        return item

    def remove(self, item):
        index = self.index(item)
        self.qt.removeAction(self.qt.actions()[index])

    def _get_action_or_menu(self, action):
        if hasattr(action, 'reflection'):
            return action.reflection()
        return action.menu().reflection()

    def __getitem__(self, index):
        action = self.qt.actions()[index]
        return self._get_action_or_menu(action)

    def __iter__(self):
        return iter(self._get_action_or_menu(action) for action in self.qt.actions())

    def __len__(self):
        return len(self.qt.actions())

    @property
    def orientation_changed(self):
        return self.__orientation_changed

    @orientation_changed.setter
    def orientation_changed(self, value):
        self.__orientation_changed = value

    def __handle_orientation_changed(self, value):
        if callable(self.orientation_changed):
            self.orientation_changed(value)



from .qutie import QtWidgets

from .frame import Frame
from .pixmap import Pixmap

__all__ = ['Label']

class Label(Frame):
    """A label widget.

    >>> label = Label('lorem')
    >>> label.text
    'lorem'

    Use a pixmap instead of text.

    >>> label = Label(pixmap='sample.png')
    """

    QtClass = QtWidgets.QLabel

    def __init__(self, text=None, *, indent=None, margin=None, pixmap=None,
                 word_wrap=None, **kwargs):
        super().__init__(**kwargs)
        if text is not None:
            self.text = text
        if indent is not None:
            self.indent = indent
        if margin is not None:
            self.margin = margin
        if pixmap is not None:
            self.pixmap = pixmap
        if word_wrap is not None:
            self.word_wrap = word_wrap

    @property
    def text(self):
        return self.qt.text()

    @text.setter
    def text(self, value):
        self.qt.setText(format(value))

    @property
    def indent(self):
        return self.qt.indent()

    @indent.setter
    def indent(self, value):
        self.qt.setIndent(int(value))

    @property
    def margin(self):
        return self.qt.margin()

    @margin.setter
    def margin(self, value):
        self.qt.setMargin(int(value))

    @property
    def pixmap(self):
        pixmap = self.qt.pixmap()
        if pixmap is not None:
            return Pixmap(qt=pixmap)
        return None

    @pixmap.setter
    def pixmap(self, value):
        if not isinstance(value, Pixmap):
            value = Pixmap(value)
        self.qt.setPixmap(value.qt)

    @property
    def word_wrap(self):
        """Enable automatic word wrap."""
        return self.qt.wordWrap()

    @word_wrap.setter
    def word_wrap(self, value):
        self.qt.setWordWrap(bool(value))

    @property
    def selected_text(self):
        return self.qt.selectedText()

    def clear(self):
        """Clears label contents."""
        self.qt.clear()



import os
from setuptools import setup, find_packages

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as f:
    long_description = f.read()

setup(
    name='qutie',
    version='1.6.1',
    author="Bernhard Arnold",
    author_email="bernhard.arnold@oeaw.ac.at",
    description="Yet another pythonic UI library using PyQt5",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/arnobaer/qutie',
    packages=find_packages(exclude=['tests']),
    install_requires=[
        'PyQt5>=5.12',
        'appdirs>=1.4.4'
    ],
    test_suite='tests',
    license="GPLv3",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3'
    ]
)



from PyQt5.QtCore import Qt
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

__all__ = ['Qutie']

def handle_event(event):
    def handle_event(*args, **kwargs):
        if callable(event):
            event(*args, **kwargs)
    return handle_event

def to_brush(method):
    def to_brush(self, value):
        if value is None:
            brush = QtGui.QBrush()
        else:
            brush = QtGui.QBrush(QtGui.QColor(value))
        return method(self, brush)
    return to_brush

def from_brush(method):
    def from_brush(self):
        value = method(self)
        if value is not None:
            return QtGui.QBrush(value).color().name()
        return None
    return from_brush

class QutieStub:

    QtClass = NotImplemented

    def __init__(self, qt=None):
        class QtClassWrapper(self.QtClass):

            def reflection(self):
                return self._reflection

            def setReflection(self, value):
                self._reflection = value

        if qt is None: qt = []
        if not isinstance(qt, (tuple, list)):
            qt=[qt]
        self.__qt = QtClassWrapper(*qt)
        self.qt.setReflection(self)

    @property
    def qt(self):
        return self.__qt

class Qutie:

    QtClass = NotImplemented

    def __init__(self, qt=None):
        class QtClassWrapper(self.QtClass):

            handleEvent = QtCore.pyqtSignal(object)

            def reflection(self):
                return self.property('__qutie_property')

            def setReflection(self, value):
                self.setProperty('__qutie_property', value)

        if qt is None: qt = []
        if not isinstance(qt, (tuple, list)):
            qt=[qt]
        self.__qt = QtClassWrapper(*qt)
        self.qt.setReflection(self)

    @property
    def qt(self):
        return self.__qt
